from __future__ import unicode_literals

import os
import pkg_resources


# -- Project settings
project = 'Picobox'
copyright = '2017, Ihor Kalnytskyi'
release = pkg_resources.get_distribution('picobox').version
version = '.'.join(release.split('.')[:2])

# -- General settings
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx']
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['_build', '_themes']
pygments_style = 'sphinx'
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}
autodoc_member_order = 'bysource'
autodoc_mock_imports = ['contextvars', 'flask']

# -- HTML output
html_use_index = False
html_show_sourcelink = False
html_logo = '_static/picobox.svg'

if not os.environ.get('READTHEDOCS') == 'True':
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Sphinx does not support "code" directive preserving default (docutils)
# behaviour. This means code won't be highlighted which is not good. In
# order to fix that we want to register custom (Sphinx) translator for
# code directive.
#
# See https://github.com/sphinx-doc/sphinx/issues/2155 for details.
from docutils.parsers.rst import directives
from sphinx.directives.code import CodeBlock
directives.register_directive('code', CodeBlock)

# flake8: noqa



"""Test picobox class."""

import collections
import inspect
import itertools
import traceback

import pytest
import picobox

from picobox import _compat


def test_box_put_key(boxclass, supported_key):
    testbox = boxclass()
    testbox.put(supported_key, 'the-value')

    assert testbox.get(supported_key) == 'the-value'


def test_box_put_value(boxclass, supported_value):
    testbox = boxclass()
    testbox.put('the-key', supported_value)

    assert testbox.get('the-key') is supported_value


def test_box_put_factory(boxclass):
    testbox = boxclass()
    testbox.put('the-key', factory=object)

    objects = [testbox.get('the-key') for _ in range(10)]

    assert len(objects) == 10
    assert len(set(map(id, objects))) == 10


def test_box_put_factory_singleton_scope(boxclass):
    testbox = boxclass()
    testbox.put('the-key', factory=object, scope=picobox.singleton)

    objects = [testbox.get('the-key') for _ in range(10)]

    assert len(objects) == 10
    assert len(set(map(id, objects))) == 1


def test_box_put_factory_custom_scope(boxclass):
    class namespacescope(picobox.Scope):

        def __init__(self):
            self._store = collections.defaultdict(dict)

        def set(self, key, value):
            self._store[namespace][key] = value

        def get(self, key):
            return self._store[namespace][key]

    testbox = boxclass()
    testbox.put('the-key', factory=object, scope=namespacescope)

    objects = []

    namespace = 'one'
    objects.extend([
        testbox.get('the-key'),
        testbox.get('the-key'),
    ])

    namespace = 'two'
    objects.extend([
        testbox.get('the-key'),
        testbox.get('the-key'),
    ])

    assert len(objects) == 4
    assert len(set(map(id, objects[:2]))) == 1
    assert len(set(map(id, objects[2:]))) == 1
    assert len(set(map(id, objects))) == 2


def test_box_put_factory_dependency(boxclass):
    testbox = boxclass()

    @testbox.pass_('a')
    def fn(a):
        return a + 1

    testbox.put('a', 13)
    testbox.put('b', factory=fn)

    assert testbox.get('b') == 14


def test_box_put_value_and_factory(boxclass):
    testbox = boxclass()

    with pytest.raises(ValueError) as excinfo:
        testbox.put('the-key', 42, factory=object)
    excinfo.match("either 'value' or 'factory'/'scope' pair must be passed")


def test_box_put_value_and_scope(boxclass):
    testbox = boxclass()

    with pytest.raises(ValueError) as excinfo:
        testbox.put('the-key', 42, scope=picobox.threadlocal)
    excinfo.match("either 'value' or 'factory'/'scope' pair must be passed")


def test_box_get_keyerror(boxclass):
    testbox = boxclass()

    with pytest.raises(KeyError, match='the-key'):
        testbox.get('the-key')


def test_box_get_default(boxclass):
    testbox = boxclass()
    sentinel = object()

    assert testbox.get('the-key', sentinel) is sentinel


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'b': 2, 'c': 3},                15),
])
def test_box_pass_a(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('a', 10)

    @testbox.pass_('a')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'c': 3},                        14),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'c': 3},                14),
])
def test_box_pass_b(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('b', 10)

    @testbox.pass_('b')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                              13),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                        13),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},                13),
])
def test_box_pass_c(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('c', 10)

    @testbox.pass_('c')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                              13),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                        13),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},                13),
])
def test_box_pass_c_default(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('c', 10)

    @testbox.pass_('c')
    def fn(a, b, c=20):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'c': 3},                       104),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'c': 3},               104),
    ((),            {'b': 2, 'c': 3},                15),
    ((),            {'c': 3},                       113),
])
def test_box_pass_ab(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('a', 10)
    testbox.put('b', 100)

    @testbox.pass_('a')
    @testbox.pass_('b')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                             103),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                       103),
    ((1,),          {'c': 3},                        14),
    ((1,),          {},                             111),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},               103),
    ((),            {'a': 1, 'c': 3},                14),
    ((),            {'a': 1},                       111),
])
def test_box_pass_bc(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('b', 10)
    testbox.put('c', 100)

    @testbox.pass_('b')
    @testbox.pass_('c')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                             103),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                       103),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},               103),
    ((),            {'b': 2, 'c': 3},                15),
    ((),            {'b': 2},                       112),
])
def test_box_pass_ac(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('a', 10)
    testbox.put('c', 100)

    @testbox.pass_('a')
    @testbox.pass_('c')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                            1003),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                      1003),
    ((1,),          {'c': 3},                       104),
    ((1,),          {},                            1101),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},              1003),
    ((),            {'a': 1, 'c': 3},               104),
    ((),            {'a': 1},                      1101),
    ((),            {},                            1110),
])
def test_box_pass_abc(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('a', 10)
    testbox.put('b', 100)
    testbox.put('c', 1000)

    @testbox.pass_('a')
    @testbox.pass_('b')
    @testbox.pass_('c')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'c': 3},                        14),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'c': 3},                14),
])
def test_box_pass_d_as_b(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('d', 10)

    @testbox.pass_('d', as_='b')
    def fn(a, b, c):
        return a + b + c

    assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, ),         {},                               1),
    ((),            {'x': 1},                         1),
    ((),            {},                              42),
])
def test_box_pass_method(args, kwargs, rv, boxclass):
    testbox = boxclass()
    testbox.put('x', 42)

    class Foo:
        @testbox.pass_('x')
        def __init__(self, x):
            self.x = x

    assert Foo(*args, **kwargs).x == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((0, ),         {},                              41),
    ((),            {'x': 0},                        41),
    ((),            {},                              42),
])
def test_box_pass_key_type(args, kwargs, rv, boxclass):
    class key:
        pass

    testbox = boxclass()
    testbox.put(key, 1)

    @testbox.pass_(key, as_='x')
    def fn(x):
        return x + 41

    assert fn(*args, **kwargs) == rv


def test_box_pass_unexpected_argument(boxclass):
    testbox = boxclass()
    testbox.put('d', 10)

    @testbox.pass_('d')
    def fn(a, b):
        return a + b

    with pytest.raises(TypeError) as excinfo:
        fn(1, 2)

    assert str(excinfo.value) == "fn() got an unexpected keyword argument 'd'"


def test_box_pass_keyerror(boxclass):
    testbox = boxclass()

    @testbox.pass_('b')
    def fn(a, b):
        return a + b

    with pytest.raises(KeyError) as excinfo:
        fn(1)

    excinfo.match('b')


def test_box_pass_optimization(boxclass, request):
    testbox = boxclass()
    testbox.put('a', 1)
    testbox.put('b', 1)
    testbox.put('d', 1)

    @testbox.pass_('a')
    @testbox.pass_('b')
    @testbox.pass_('d', as_='c')
    def fn(a, b, c):
        backtrace = list(itertools.dropwhile(
            lambda frame: frame[2] != request.function.__name__,
            traceback.extract_stack()))
        return backtrace[1:-1]

    assert len(fn()) == 1


def test_box_pass_optimization_complex(boxclass, request):
    testbox = boxclass()
    testbox.put('a', 1)
    testbox.put('b', 1)
    testbox.put('c', 1)
    testbox.put('d', 1)

    def passthrough(fn):
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)
        return wrapper

    @testbox.pass_('a')
    @testbox.pass_('b')
    @passthrough
    @testbox.pass_('c')
    @testbox.pass_('d')
    def fn(a, b, c, d):
        backtrace = list(itertools.dropwhile(
            lambda frame: frame[2] != request.function.__name__,
            traceback.extract_stack()))
        return backtrace[1:-1]

    assert len(fn()) == 3


def test_chainbox_put_changes_box():
    testbox = picobox.Box()
    testchainbox = picobox.ChainBox(testbox)

    with pytest.raises(KeyError, match='the-key'):
        testchainbox.get('the-key')
    testchainbox.put('the-key', 42)

    assert testbox.get('the-key') == 42


def test_chainbox_get_chained():
    testbox_a = picobox.Box()
    testbox_a.put('the-key', 42)

    testbox_b = picobox.Box()
    testbox_b.put('the-key', 13)
    testbox_b.put('the-pin', 12)

    testchainbox = picobox.ChainBox(testbox_a, testbox_b)

    assert testchainbox.get('the-key') == 42
    assert testchainbox.get('the-pin') == 12


def test_chainbox_isinstance_box():
    assert isinstance(picobox.ChainBox(), picobox.Box)


@pytest.mark.parametrize('name', [
    name
    for name, _ in inspect.getmembers(picobox.Box) if not name.startswith('_')
])
def test_chainbox_box_interface(name):
    boxsignature = _compat.signature(getattr(picobox.Box(), name))
    chainboxsignature = _compat.signature(getattr(picobox.ChainBox(), name))

    assert boxsignature == chainboxsignature



"""Setup pytest environment."""

import pytest
import picobox


@pytest.fixture(scope='function', params=[
    42,
    '42',
    42.42,
    True,
    None,
    (1, None, True),
    object(),
])
def hashable_value(request):
    return request.param


@pytest.fixture(scope='function', params=[
    42,
    '42',
    42.42,
    True,
    None,
    {'a': 1, 'b': 2},
    {'a', 'b', 'c'},
    [1, 2, 'c'],
    (1, None, True),
    object(),
    lambda: 42,
])
def any_value(request):
    return request.param


@pytest.fixture(scope='function')
def supported_key(hashable_value):
    return hashable_value


@pytest.fixture(scope='function')
def supported_value(any_value):
    return any_value


@pytest.fixture(params=[picobox.Box, picobox.ChainBox])
def boxclass(request):
    yield request.param



import picobox


@picobox.pass_('conf')
def session(conf):
    class Session:
        connection = conf['connection']
    return Session()


@picobox.pass_('session')
def compute(session):
    print(session.connection)


box = picobox.Box()
box.put('conf', {'connection': 'sqlite://'})
box.put('session', factory=session)

with picobox.push(box):
    compute()



"""Test Flask scopes."""

import flask
import pytest

from picobox.contrib import flaskscopes


@pytest.fixture(scope='function')
def appscope():
    return flaskscopes.application()


@pytest.fixture(scope='function')
def reqscope():
    return flaskscopes.request()


@pytest.fixture(scope='function')
def flaskapp():
    return flask.Flask('test')


@pytest.fixture(scope='function')
def appcontext(flaskapp):
    return flaskapp.app_context


@pytest.fixture(scope='function')
def reqcontext(flaskapp):
    return flaskapp.test_request_context


@pytest.mark.parametrize('scopename, ctx', [
    ('appscope', 'appcontext'),
    ('reqscope', 'reqcontext'),
])
def test_scope_set(request, scopename, ctx, supported_key, supported_value):
    scope = request.getfixturevalue(scopename)
    ctxfn = request.getfixturevalue(ctx)

    with ctxfn():
        scope.set(supported_key, supported_value)
        assert scope.get(supported_key) is supported_value


@pytest.mark.parametrize('scopename', [
    'appscope',
    'reqscope',
])
def test_scope_set_nocontext(request, scopename):
    scope = request.getfixturevalue(scopename)

    with pytest.raises(RuntimeError) as excinfo:
        scope.set('the-key', 'the-value')
    excinfo.match('Working outside of application context.')


@pytest.mark.parametrize('scopename, ctx', [
    ('appscope', 'appcontext'),
    ('reqscope', 'reqcontext'),
])
def test_scope_set_value_overwrite(request, scopename, ctx):
    scope = request.getfixturevalue(scopename)
    ctxfn = request.getfixturevalue(ctx)
    value = object()

    with ctxfn():
        scope.set('the-key', value)
        assert scope.get('the-key') is value

        scope.set('the-key', 'overwrite')
        assert scope.get('the-key') == 'overwrite'


@pytest.mark.parametrize('scopename, ctx', [
    ('appscope', 'appcontext'),
    ('reqscope', 'reqcontext'),
])
def test_scope_get_keyerror(request, scopename, ctx, supported_key):
    scope = request.getfixturevalue(scopename)
    ctxfn = request.getfixturevalue(ctx)

    with pytest.raises(KeyError, match=repr(supported_key)):
        with ctxfn():
            scope.get(supported_key)


@pytest.mark.parametrize('scopename, ctx', [
    ('appscope', 'appcontext'),
    ('reqscope', 'reqcontext'),
])
def test_scope_state_not_leaked(request, scopename, ctx):
    ctxfn = request.getfixturevalue(ctx)

    scope_a = request.getfixturevalue(scopename)
    value_a = object()

    scope_b = type(scope_a)()
    value_b = object()

    with ctxfn():
        scope_a.set('the-key', value_a)
        assert scope_a.get('the-key') is value_a

        with pytest.raises(KeyError, match='the-key'):
            scope_b.get('the-key')

        scope_b.set('the-key', value_b)
        assert scope_b.get('the-key') is value_b

        scope_a.set('the-key', value_a)
        assert scope_a.get('the-key') is value_a


@pytest.mark.parametrize('scopename, ctx', [
    ('appscope', 'appcontext'),
])
def test_scope_value_shared(request, scopename, ctx):
    scope = request.getfixturevalue(scopename)
    ctxfn = request.getfixturevalue(ctx)
    value = object()

    with ctxfn():
        scope.set('the-key', value)

    with ctxfn():
        assert scope.get('the-key') is value


@pytest.mark.parametrize('scopename, ctx', [
    ('reqscope', 'reqcontext'),
])
def test_scope_value_not_shared(request, scopename, ctx):
    scope = request.getfixturevalue(scopename)
    ctxfn = request.getfixturevalue(ctx)

    with ctxfn():
        scope.set('the-key', 'the-value')

    with pytest.raises(KeyError, match='the-key'):
        with ctxfn():
            scope.get('the-key')


def test_scope_value_not_shared_between_apps(appscope):
    eggs = flask.Flask('eggs')
    rice = flask.Flask('rice')
    value = object()

    with eggs.app_context():
        appscope.set('the-key', value)

    with eggs.app_context():
        assert appscope.get('the-key') is value

    with pytest.raises(KeyError, match='the-key'):
        with rice.app_context():
            appscope.get('the-key')

    with eggs.app_context():
        assert appscope.get('the-key') is value



"""Dependency injection framework designed with Python in mind."""

from ._box import Box, ChainBox
from ._scopes import Scope, singleton, threadlocal, noscope
from ._stack import Stack, push, pop, put, get, pass_

try:
    from ._scopes import contextvars
except ImportError:
    pass


__all__ = [
    'Box',
    'ChainBox',

    'Scope',
    'singleton',
    'threadlocal',
    'contextvars',
    'noscope',

    'Stack',
    'push',
    'pop',
    'put',
    'get',
    'pass_',
]



"""Test picobox's stack interface."""

import itertools
import traceback

import pytest
import picobox


@pytest.fixture(params=[picobox.Stack(), picobox])
def teststack(request):
    yield request.param


def test_box_put_key(boxclass, teststack, supported_key):
    testbox = boxclass()

    with teststack.push(testbox):
        teststack.put(supported_key, 'the-value')

    assert testbox.get(supported_key) == 'the-value'


def test_box_put_value(boxclass, teststack, supported_value):
    testbox = boxclass()

    with teststack.push(testbox):
        teststack.put('the-key', supported_value)

    assert testbox.get('the-key') is supported_value


def test_box_put_factory(boxclass, teststack):
    testbox = boxclass()

    with teststack.push(testbox):
        teststack.put('the-key', factory=object)

    objects = [testbox.get('the-key') for _ in range(10)]

    assert len(objects) == 10
    assert len(set(map(id, objects))) == 10


def test_box_put_factory_singleton_scope(boxclass, teststack):
    testbox = boxclass()

    with teststack.push(testbox):
        teststack.put('the-key', factory=object, scope=picobox.singleton)

    objects = [testbox.get('the-key') for _ in range(10)]

    assert len(objects) == 10
    assert len(set(map(id, objects))) == 1


def test_box_put_factory_dependency(boxclass, teststack):
    testbox = boxclass()

    @teststack.pass_('a')
    def fn(a):
        return a + 1

    with teststack.push(testbox):
        teststack.put('a', 13)
        teststack.put('b', factory=fn)

        assert teststack.get('b') == 14


def test_box_put_value_and_factory(boxclass, teststack):
    testbox = boxclass()

    with teststack.push(testbox):
        with pytest.raises(ValueError) as excinfo:
            teststack.put('the-key', 42, factory=object)
    excinfo.match("either 'value' or 'factory'/'scope' pair must be passed")


def test_box_put_value_and_scope(boxclass, teststack):
    testbox = boxclass()

    with teststack.push(testbox):
        with pytest.raises(ValueError) as excinfo:
            teststack.put('the-key', 42, scope=picobox.threadlocal)
    excinfo.match("either 'value' or 'factory'/'scope' pair must be passed")


def test_box_put_runtimeerror(boxclass, teststack):
    with pytest.raises(RuntimeError) as excinfo:
        teststack.put('the-key', object())

    assert str(excinfo.value) == (
        'No boxes found on the stack, please `.push()` a box first.')


def test_box_get_value(boxclass, teststack, supported_value):
    testbox = boxclass()
    testbox.put('the-key', supported_value)

    with teststack.push(testbox):
        assert teststack.get('the-key') is supported_value


def test_box_get_keyerror(boxclass, teststack):
    testbox = boxclass()

    with teststack.push(testbox):
        with pytest.raises(KeyError, match='the-key'):
            teststack.get('the-key')


def test_box_get_default(boxclass, teststack):
    testbox = boxclass()
    sentinel = object()

    with teststack.push(testbox):
        assert teststack.get('the-key', sentinel) is sentinel


@pytest.mark.parametrize('kwargs', [
    {},
    {'chain': False},
])
def test_box_get_from_top(boxclass, teststack, kwargs):
    testbox_a = boxclass()
    testbox_a.put('the-key', 'a')
    testbox_a.put('the-pin', 'a')

    testbox_b = boxclass()
    testbox_b.put('the-key', 'b')

    with teststack.push(testbox_a):
        assert teststack.get('the-key') == 'a'
        assert teststack.get('the-pin') == 'a'

        with teststack.push(testbox_b, **kwargs):
            assert teststack.get('the-key') == 'b'

            with pytest.raises(KeyError, match='the-pin'):
                teststack.get('the-pin')

        assert teststack.get('the-key') == 'a'
        assert teststack.get('the-pin') == 'a'


def test_box_get_from_top_chain(boxclass, teststack):
    testbox_a = boxclass()
    testbox_a.put('the-key', 'a')
    testbox_a.put('the-pin', 'a')

    testbox_b = boxclass()
    testbox_b.put('the-key', 'b')

    testbox_c = boxclass()
    testbox_c.put('the-tip', 'c')

    with teststack.push(testbox_a):
        assert teststack.get('the-key') == 'a'
        assert teststack.get('the-pin') == 'a'

        with teststack.push(testbox_b, chain=True):
            assert teststack.get('the-key') == 'b'
            assert teststack.get('the-pin') == 'a'

            with teststack.push(testbox_c, chain=True):
                assert teststack.get('the-tip') == 'c'
                assert teststack.get('the-key') == 'b'
                assert teststack.get('the-pin') == 'a'


def test_box_get_runtimeerror(teststack):
    with pytest.raises(RuntimeError) as excinfo:
        teststack.get('the-key')

    assert str(excinfo.value) == (
        'No boxes found on the stack, please `.push()` a box first.')


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'b': 2, 'c': 3},                15),
])
def test_box_pass_a(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('a', 10)

    @teststack.pass_('a')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'c': 3},                        14),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'c': 3},                14),
])
def test_box_pass_b(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('b', 10)

    @teststack.pass_('b')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                              13),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                        13),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},                13),
])
def test_box_pass_c(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('c', 10)

    @teststack.pass_('c')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                              13),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                        13),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},                13),
])
def test_box_pass_c_default(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('c', 10)

    @teststack.pass_('c')
    def fn(a, b, c=20):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'c': 3},                       104),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'c': 3},               104),
    ((),            {'b': 2, 'c': 3},                15),
    ((),            {'c': 3},                       113),
])
def test_box_pass_ab(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('a', 10)
    testbox.put('b', 100)

    @teststack.pass_('a')
    @teststack.pass_('b')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                             103),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                       103),
    ((1,),          {'c': 3},                        14),
    ((1,),          {},                             111),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},               103),
    ((),            {'a': 1, 'c': 3},                14),
    ((),            {'a': 1},                       111),
])
def test_box_pass_bc(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('b', 10)
    testbox.put('c', 100)

    @teststack.pass_('b')
    @teststack.pass_('c')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                             103),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                       103),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},               103),
    ((),            {'b': 2, 'c': 3},                15),
    ((),            {'b': 2},                       112),
])
def test_box_pass_ac(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('a', 10)
    testbox.put('c', 100)

    @teststack.pass_('a')
    @teststack.pass_('c')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1, 2),        {},                            1003),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'b': 2},                      1003),
    ((1,),          {'c': 3},                       104),
    ((1,),          {},                            1101),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'b': 2},              1003),
    ((),            {'a': 1, 'c': 3},               104),
    ((),            {'a': 1},                      1101),
    ((),            {},                            1110),
])
def test_box_pass_abc(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('a', 10)
    testbox.put('b', 100)
    testbox.put('c', 1000)

    @teststack.pass_('a')
    @teststack.pass_('b')
    @teststack.pass_('c')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, 2, 3),     {},                               6),
    ((1, 2),        {'c': 3},                         6),
    ((1,),          {'b': 2, 'c': 3},                 6),
    ((1,),          {'c': 3},                        14),
    ((),            {'a': 1, 'b': 2, 'c': 3},         6),
    ((),            {'a': 1, 'c': 3},                14),
])
def test_box_pass_d_as_b(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('d', 10)

    @teststack.pass_('d', as_='b')
    def fn(a, b, c):
        return a + b + c

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((1, ),         {},                               1),
    ((),            {'x': 1},                         1),
    ((),            {},                              42),
])
def test_box_pass_method(boxclass, teststack, args, kwargs, rv):
    testbox = boxclass()
    testbox.put('x', 42)

    class Foo:
        @teststack.pass_('x')
        def __init__(self, x):
            self.x = x

    with teststack.push(testbox):
        assert Foo(*args, **kwargs).x == rv


@pytest.mark.parametrize('args, kwargs, rv', [
    ((0, ),         {},                              41),
    ((),            {'x': 0},                        41),
    ((),            {},                              42),
])
def test_box_pass_key_type(boxclass, teststack, args, kwargs, rv):
    class key:
        pass

    testbox = boxclass()
    testbox.put(key, 1)

    @testbox.pass_(key, as_='x')
    def fn(x):
        return x + 41

    with teststack.push(testbox):
        assert fn(*args, **kwargs) == rv


def test_box_pass_unexpected_argument(boxclass, teststack):
    testbox = boxclass()
    testbox.put('d', 10)

    @teststack.pass_('d')
    def fn(a, b):
        return a + b

    with teststack.push(testbox):
        with pytest.raises(TypeError) as excinfo:
            fn(1, 2)

    assert str(excinfo.value) == "fn() got an unexpected keyword argument 'd'"


def test_box_pass_keyerror(boxclass, teststack):
    testbox = boxclass()

    @teststack.pass_('b')
    def fn(a, b):
        return a + b

    with teststack.push(testbox):
        with pytest.raises(KeyError, match='b'):
            fn(1)


def test_box_pass_runtimeerror(teststack):
    @teststack.pass_('a')
    def fn(a):
        return a

    with pytest.raises(RuntimeError) as excinfo:
        fn()

    assert str(excinfo.value) == (
        'No boxes found on the stack, please `.push()` a box first.')


def test_box_pass_optimization(boxclass, teststack, request):
    testbox = boxclass()
    testbox.put('a', 1)
    testbox.put('b', 1)
    testbox.put('d', 1)

    @teststack.pass_('a')
    @teststack.pass_('b')
    @teststack.pass_('d', as_='c')
    def fn(a, b, c):
        backtrace = list(itertools.dropwhile(
            lambda frame: frame[2] != request.function.__name__,
            traceback.extract_stack()))
        return backtrace[1:-1]

    with teststack.push(testbox):
        assert len(fn()) == 1


def test_box_pass_optimization_complex(boxclass, teststack, request):
    testbox = boxclass()
    testbox.put('a', 1)
    testbox.put('b', 1)
    testbox.put('c', 1)
    testbox.put('d', 1)

    def passthrough(fn):
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)
        return wrapper

    @teststack.pass_('a')
    @teststack.pass_('b')
    @passthrough
    @teststack.pass_('c')
    @teststack.pass_('d')
    def fn(a, b, c, d):
        backtrace = list(itertools.dropwhile(
            lambda frame: frame[2] != request.function.__name__,
            traceback.extract_stack()))
        return backtrace[1:-1]

    with teststack.push(testbox):
        assert len(fn()) == 3


def test_chainbox_put_changes_box(teststack):
    testbox = picobox.Box()
    testchainbox = picobox.ChainBox(testbox)

    with teststack.push(testchainbox):
        with pytest.raises(KeyError, match='the-key'):
            teststack.get('the-key')
        teststack.put('the-key', 42)

        assert testbox.get('the-key') == 42


def test_chainbox_get_chained(teststack):
    testbox_a = picobox.Box()
    testbox_a.put('the-key', 42)

    testbox_b = picobox.Box()
    testbox_b.put('the-key', 13)
    testbox_b.put('the-pin', 12)

    testchainbox = picobox.ChainBox(testbox_a, testbox_b)

    with teststack.push(testchainbox):
        assert testchainbox.get('the-key') == 42
        assert testchainbox.get('the-pin') == 12


def test_stack_isolated(boxclass):
    testbox_a = picobox.Box()
    testbox_a.put('the-key', 42)

    testbox_b = picobox.Box()
    testbox_b.put('the-pin', 12)

    teststack_a = picobox.Stack()
    teststack_b = picobox.Stack()

    with teststack_a.push(testbox_a):
        with pytest.raises(RuntimeError) as excinfo:
            teststack_b.get('the-key')
        assert str(excinfo.value) == (
            'No boxes found on the stack, please `.push()` a box first.')

        with teststack_b.push(testbox_b):
            with pytest.raises(KeyError, match='the-pin'):
                teststack_a.get('the-pin')
            assert teststack_b.get('the-pin') == 12


def test_push_pop_as_regular_functions(teststack):
    @teststack.pass_('magic')
    def do(magic):
        return magic + 1

    foobox = picobox.Box()
    foobox.put('magic', 42)

    barbox = picobox.Box()
    barbox.put('magic', 13)

    teststack.push(foobox)
    assert do() == 43

    teststack.push(barbox)
    assert do() == 14

    assert teststack.pop() is barbox
    assert teststack.pop() is foobox


def test_pop_empty_stack(teststack):
    with pytest.raises(IndexError):
        teststack.pop()



"""Test picobox's scopes implementations."""

import threading

import pytest
import picobox


@pytest.fixture(scope='function')
def singleton():
    return picobox.singleton()


@pytest.fixture(scope='function')
def threadlocal():
    return picobox.threadlocal()


@pytest.fixture(scope='function')
def contextvars():
    pytest.importorskip('contextvars')
    return picobox.contextvars()


@pytest.fixture(scope='function')
def noscope():
    return picobox.noscope()


@pytest.fixture(scope='function')
def exec_thread():
    """Run a given callback in a separate OS thread."""
    def executor(callback, *args, **kwargs):
        closure = {}

        def target():
            try:
                closure['ret'] = callback(*args, **kwargs)
            except Exception as e:
                closure['exc'] = e

        worker = threading.Thread(target=target)
        worker.start()
        worker.join()

        if 'exc' in closure:
            raise closure['exc']
        return closure['ret']
    return executor


@pytest.fixture(scope='function')
def exec_coroutine(request):
    """Run a given coroutine function in a separate event loop."""
    asyncio = pytest.importorskip('asyncio')
    loop = asyncio.new_event_loop()
    request.addfinalizer(loop.close)

    def executor(coroutine_function, *args, **kwargs):
        if not asyncio.iscoroutinefunction(coroutine_function):
            coroutine_function = asyncio.coroutine(coroutine_function)
        return loop.run_until_complete(coroutine_function(*args, **kwargs))
    return executor


@pytest.fixture(scope='function')
def exec_context():
    """Run a given callback in a separate context (PEP 567)."""
    def executor(callback, *args, **kwargs):
        import contextvars
        context = contextvars.copy_context()
        return context.run(callback, *args, **kwargs)

    pytest.importorskip('contextvars')
    return executor


def test_scope_contextvars_attribute_error(monkeypatch):
    try:
        __import__('contextvars')
        pytest.skip("could import 'contextvars'")
    except ImportError:
        pass

    with pytest.raises(AttributeError, match="has no attribute 'contextvars'"):
        picobox.contextvars


@pytest.mark.parametrize('scopename', [
    'singleton',
    'threadlocal',
    'contextvars',
])
def test_scope_set(request, scopename, supported_key, supported_value):
    scope = request.getfixturevalue(scopename)

    scope.set(supported_key, supported_value)
    assert scope.get(supported_key) is supported_value


def test_scope_set_noscope(supported_key, supported_value):
    scope = picobox.noscope()
    scope.set(supported_key, supported_value)

    with pytest.raises(KeyError, match=str(supported_key)):
        scope.get(supported_key)


@pytest.mark.parametrize('scopename', [
    'singleton',
    'threadlocal',
    'contextvars',
])
def test_scope_set_overwrite(request, scopename):
    scope = request.getfixturevalue(scopename)
    value = object()

    scope.set('the-key', value)
    assert scope.get('the-key') is value

    scope.set('the-key', 'overwrite')
    assert scope.get('the-key') == 'overwrite'


@pytest.mark.parametrize('scopename', [
    'singleton',
    'threadlocal',
    'contextvars',
    'noscope',
])
def test_scope_get_keyerror(request, scopename, supported_key):
    scope = request.getfixturevalue(scopename)

    with pytest.raises(KeyError, match=repr(supported_key)):
        scope.get(supported_key)


@pytest.mark.parametrize('scopename', [
    'singleton',
    'threadlocal',
    'contextvars',
])
def test_scope_state_not_leaked(request, scopename):
    scope_a = request.getfixturevalue(scopename)
    value_a = object()

    scope_b = type(scope_a)()
    value_b = object()

    scope_a.set('the-key', value_a)
    assert scope_a.get('the-key') is value_a

    with pytest.raises(KeyError, match='the-key'):
        scope_b.get('the-key')

    scope_b.set('the-key', value_b)
    assert scope_b.get('the-key') is value_b

    scope_a.set('the-key', value_a)
    assert scope_a.get('the-key') is value_a


@pytest.mark.parametrize('scopename, executor', [
    ('singleton',    'exec_thread'),
    ('singleton',    'exec_coroutine'),
    ('singleton',    'exec_context'),
    ('threadlocal',  'exec_coroutine'),
    ('threadlocal',  'exec_context'),
])
def test_scope_value_shared(request, scopename, executor):
    scope = request.getfixturevalue(scopename)
    value = object()
    exec_ = request.getfixturevalue(executor)

    exec_(scope.set, 'the-key', value)
    assert exec_(scope.get, 'the-key') is value


@pytest.mark.parametrize('scopename, executor', [
    ('threadlocal',  'exec_thread'),
    ('contextvars',  'exec_thread'),
    ('contextvars',  'exec_coroutine'),
    ('contextvars',  'exec_context'),
    ('noscope',      'exec_thread'),
    ('noscope',      'exec_coroutine'),
    ('noscope',      'exec_context'),
])
def test_scope_value_not_shared(request, scopename, executor):
    scope = request.getfixturevalue(scopename)
    value = object()
    exec_ = request.getfixturevalue(executor)

    exec_(scope.set, 'the-key', value)

    with pytest.raises(KeyError, match='the-key'):
        exec_(scope.get, 'the-key')


@pytest.mark.parametrize('scopename, executor', [
    ('singleton',    'exec_thread'),
    ('singleton',    'exec_coroutine'),
    ('singleton',    'exec_context'),
    ('threadlocal',  'exec_thread'),
    ('threadlocal',  'exec_coroutine'),
    ('threadlocal',  'exec_context'),
    ('contextvars',  'exec_thread'),
    ('contextvars',  'exec_coroutine'),
    ('contextvars',  'exec_context'),
])
def test_scope_value_downstack_shared(request, scopename, executor):
    scope = request.getfixturevalue(scopename)
    value = object()
    exec_ = request.getfixturevalue(executor)

    def caller():
        scope.set('the-key', value)
        return callee()

    def callee():
        return scope.get('the-key')

    assert exec_(caller) is value


@pytest.mark.parametrize('scopename, executor', [
    ('noscope',      'exec_thread'),
    ('noscope',      'exec_coroutine'),
    ('noscope',      'exec_context'),
])
def test_scope_value_downstack_not_shared(request, scopename, executor):
    scope = request.getfixturevalue(scopename)
    value = object()
    exec_ = request.getfixturevalue(executor)

    def caller():
        scope.set('the-key', value)
        return callee()

    def callee():
        return scope.get('the-key')

    with pytest.raises(KeyError, match='the-key'):
        exec_(caller)


@pytest.mark.parametrize('scopename, executor', [
    ('threadlocal',  'exec_thread'),
    ('contextvars',  'exec_thread'),
    ('contextvars',  'exec_coroutine'),
    ('contextvars',  'exec_context'),
])
def test_scope_not_leaked(request, scopename, executor):
    scope = request.getfixturevalue(scopename)
    exec_ = request.getfixturevalue(executor)

    scope.set('a-key', 'a-value')
    exec_(scope.set, 'the-key', 'the-value')

    with pytest.raises(KeyError, match='the-key'):
        scope.get('the-key')



import os
import io
import setuptools


here = os.path.dirname(__file__)

with io.open(os.path.join(here, 'README.rst'), 'r', encoding='utf-8') as f:
    long_description = f.read()


setuptools.setup(
    name='picobox',
    description='Dependency injection framework designed with Python in mind.',
    long_description=long_description,
    url='https://github.com/ikalnytskyi/picobox',
    license='MIT',
    author='Ihor Kalnytskyi',
    author_email='ihor@kalnytskyi.com',
    packages=setuptools.find_packages(where='src'),
    package_dir={'': 'src'},
    use_scm_version={
        'root': here,
    },
    setup_requires=[
        'setuptools_scm',
    ],
    install_requires=[
        'funcsigs; python_version < "3"',
    ],
    project_urls={
        'Documentation': 'https://picobox.readthedocs.io',
        'Source': 'https://github.com/ikalnytskyi/picobox',
        'Bugs': 'https://github.com/ikalnytskyi/picobox/issues',
    },
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
    ],
)



"""Py2/Py3 compatibility misc."""

import sys


if sys.version_info[0] > 2:
    signature = __import__('inspect').signature
else:
    # inspect.signature has been introduced in Python 3.3 and older versions
    # have to use third-party 'funcsigs' package instead.
    signature = __import__('funcsigs').signature


# This piece of code is copied from six library and is distributed under the
# same license (http://six.readthedocs.io/#six.add_metaclass).
def add_metaclass(metaclass):
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper
